unit Steganografia;

interface

uses Graphics, Classes;

type
  TStegoStream = class(TStream)
  private
    fBitsXColor: Integer;
    X, Y: Integer;
    ColorPos: Integer;
    BitPos: Integer;

    Width: Integer;
    Height: Integer;
    BMP: TBitmap;

    fMaxSize: Integer;
    fSize: Integer;
    HeaderPixels: Integer;

    procedure NextBit;
    procedure SetBit(Value: Integer);
    function GetBit: Integer;
    function ReadByte: Integer;
    procedure WriteByte(Value: Integer);
    function XYBC2Position: Integer;
    procedure Position2XYBC(Position: Integer);
    function BitsXPixel: Integer;
  public
    constructor Create(ABMP: TBitmap; ABitsXColor: Integer);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; overload; override;
    procedure SetSize(NewSize: Longint); overload; override;
    procedure WriteStego(const FileName: string);

    property MaxSize: Integer read fMaxSize;
    property BitsXColor: Integer read fBitsXColor;
  end;

implementation

uses Math, SysUtils;

const
  MAGIC: Integer = $01234569;

type
  TStegoHeader = record
    Magic: Integer;
    BitsXColor: Integer;
    Size: Integer;
  end;

constructor TStegoStream.Create(ABMP: TBitmap; ABitsXColor: Integer);
var
  Header: TStegoHeader;
begin
  inherited Create;
  BMP := ABMP;
  BMP.PixelFormat := pf24bit;
  fBitsXColor := 1;
  Width := BMP.Width;
  Height := BMP.Height;
  HeaderPixels := ((SizeOf(TStegoHeader) * 8 + 2) div 3);
  if (Width * Height < HeaderPixels) then 
    Raise Exception.Create('Image is too small!');
  fBitsXColor := ABitsXColor;
  if Read(Header, SizeOf(TStegoHeader)) = SizeOf(TStegoHeader) then 
    if (Header.Magic = MAGIC) then begin
      fSize := Header.Size;
      fBitsXColor := Header.BitsXColor;
    end;
  Position2XYBC(0);
  fMaxSize := ((Height * Width - HeaderPixels) * BitsXPixel) div 8;
end;

destructor TStegoStream.Destroy;
begin
  BMP.Free;
  inherited Destroy;
end;

procedure TStegoStream.NextBit;
begin
  Inc(BitPos);
  if BitPos = fBitsXColor then begin
    BitPos := 0;
    Inc(ColorPos);
    if ColorPos = 3 then begin
      ColorPos := 0;
      Inc(X);
      if X = Width then begin
        X := 0;
        Inc(Y);
      end;
    end;
  end;
end;

function TStegoStream.BitsXPixel: Integer;
begin
  Result := fBitsXColor * 3;
end;

procedure TStegoStream.SetBit(Value: Integer);
var
  NewColor: TColor;
  Pos: Integer;
begin
  NewColor := BMP.Canvas.Pixels[X, Y];
  Value := Value and 1;
  Pos := ColorPos * 8 + BitPos;
  NewColor := NewColor and not(1 shl Pos);
  NewColor := NewColor or (Value shl Pos);
  BMP.Canvas.Pixels[X, Y] := NewColor;
  NextBit;
end;

function TStegoStream.GetBit: Integer;
var
  Color: TColor;
  Pos: Integer;
begin
  Color := BMP.Canvas.Pixels[X, Y];
  Pos := ColorPos * 8 + BitPos;
  Result := Color shr Pos;
  Result := Result and 1;
  NextBit;
end;

procedure TStegoStream.WriteByte(Value: Integer);
var
  I: Integer;
begin
  for I := 0 to 7 do
    SetBit(Value shr I);
end;

function TStegoStream.ReadByte: Integer;
var
  I, J: Integer;
begin
  Result := 0;

  for I := 0 to 7 do begin
    J := GetBit;
    Result := Result or (J shl I);
  end;
end;

function TStegoStream.XYBC2Position: Integer;
begin
  Result := ((Y * Width + X - HeaderPixels) * BitsXPixel + ColorPos * fBitsXColor + BitPos) div 8;
end;

procedure TStegoStream.Position2XYBC(Position: Integer);
var
  bfb, pix, resbit: Integer;
begin
  bfb := Position * 8;
  pix := bfb div BitsXPixel + HeaderPixels;
  X := pix mod Width;
  Y := pix div Width;
  resbit := bfb mod BitsXPixel;
  ColorPos := resbit div fBitsXColor;
  BitPos := resbit mod fBitsXColor;
end;

function TStegoStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
  case Origin of
    soFromBeginning:
      begin
        Result := Offset;
      end;
    soFromCurrent:
      begin
      	Result := XYBC2Position + Offset;
      end;
    soFromEnd:
      begin
      	Result := fSize + Offset;
      end;
  else
    Result := 0;
  end;
  if (Result < 0) or (Result > fSize) then
    raise Exception.Create('Invalid offset!');
  Position2XYBC(Result);
end;

function TStegoStream.Read(var Buffer; Count: Longint): Longint;
var
  I: Integer;
begin
  Result := Min(Count, fSize - XYBC2Position);
  for I := 0 to Result - 1 do begin
    TByteArray(Buffer)[I] := ReadByte;
  end;
end;

function TStegoStream.Write(const Buffer; Count: Longint): Longint;
var
  I: Integer;
begin
  Result := Min(Count, fMaxSize - XYBC2Position);
  for I := 0 to Result - 1 do begin
    WriteByte(TByteArray(Buffer)[I]);
  end;
  fSize := Max(fSize, XYBC2Position);
end;

procedure TStegoStream.WriteStego(const FileName: string);
var
  OldX, OldY, OldC, OldB: Integer;
  Header: TStegoHeader;
begin
  OldX := X;
  OldY := Y;
  OldC := ColorPos;
  OldB := BitPos;
  X := 0;
  Y := 0;
  ColorPos := 0;
  BitPos := 0;
  Header.Magic := MAGIC;
  Header.Size := fSize;
  Header.BitsXColor := fBitsXColor;
  fBitsXColor := 1;
  Write(Header, SizeOf(TStegoHeader));
  fBitsXColor := Header.BitsXColor;
  X := OldX;
  Y := OldY;
  ColorPos := OldC;
  BitPos := OldB;
  BMP.SaveToFile(FileName);
end;

procedure TStegoStream.SetSize(NewSize: Longint);
begin
  if NewSize <= fMaxSize then begin
    fSize := NewSize;
    if XYBC2Position > fSize then
      Position2XYBC(fSize);
  end else
    Raise Exception.Create('Invalid size!');
end;

end.
